/* libgieditor
 *
 * Juno Gi Editor
 *
 * Copyright (C) 2012 Kim Taylor <kmtaylor@gmx.com>
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation version 2 of the License.
 * 
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 * 
 *	You should have received a copy of the GNU General Public License along
 *	with this program; if not, write to the Free Software Foundation, Inc.,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: libgieditor.c,v 1.5 2012/06/08 09:58:20 kmtaylor Exp $
 */

#include "libgieditor.h"
#include "midi_addresses.h"
#include "sysex.h"

#define allocate(t, num) __common_allocate(sizeof(t) * num, "libgieditor")
#define NUM_ADDRESSES libgieditor_num_addresses

int libgieditor_init(const char *client_name) {
	int retval;

	retval = sysex_init(client_name);

	return retval;
}

int libgieditor_close(void) {
	int retval;

	retval = sysex_close();

	return retval;
}

uint32_t libgieditor_add_addresses(uint32_t address1, uint32_t address2) {
	uint8_t a1, a2, a3, a4, b1, b2, b3, b4, c1, c2, c3, c4;
	a1 = (address1 & 0xff000000) >> 24; b1 = (address2 & 0xff000000) >> 24;
	a2 = (address1 & 0x00ff0000) >> 16; b2 = (address2 & 0x00ff0000) >> 16;
	a3 = (address1 & 0x0000ff00) >> 8;  b3 = (address2 & 0x0000ff00) >> 8;
	a4 = (address1 & 0x000000ff);	    b4 = (address2 & 0x000000ff);

	c4 = a4 + b4;
	c3 = a3 + b3 + ((c4 & 0x80)>>7);
	c2 = a2 + b2 + ((c3 & 0x80)>>7);
	c1 = a1 + b1 + ((c2 & 0x80)>>7);

	c4 &= 0x7f;
	c3 &= 0x7f;
	c2 &= 0x7f;
	c1 &= 0x7f;

	return ( c1<<24 | c2<<16 | c3<<8 | c4 );
}

midi_address *libgieditor_match_midi_address(uint32_t sysex_addr) {
	int i;
        for (i = 0; i < NUM_ADDRESSES; i++) {
                if (libgieditor_midi_addresses[i].sysex_addr == sysex_addr) {
                        return &libgieditor_midi_addresses[i];
                }
        }
        return NULL;
}

static uint32_t class_member_base(MidiClass *class, uint32_t sysex_addr) {
	int i;
	for (i = 0; i < class->size; i++) {
            if (class->members[i].sysex_addr_base > sysex_addr)
                return class->members[i - 1].sysex_addr_base;
        }
	return class->members[i - 1].sysex_addr_base;
}

/* Once each match is made, we can subtract its base address */
static uint32_t match_class_member(uint32_t sysex_addr, int depth) {
	int i;
	uint32_t sub_addr = 0;
	int num_parents = 0;
	MidiClass *cur_parent = NULL;
	
	MidiClass *class = libgieditor_match_midi_address(sysex_addr)->class;
	if (depth)
	    class = class->parents[depth - 1];

	if (class->parents)
	    cur_parent = class->parents[0];

	while(cur_parent) {
	    cur_parent = class->parents[num_parents];
	    num_parents++;
	}
	num_parents--;
	for (i = 0; i < num_parents; i++) {
	    cur_parent = class->parents[num_parents - 1 - i];
	    sub_addr += class_member_base(cur_parent, sysex_addr - sub_addr);
	}

	sysex_addr -= sub_addr;

	for (i = 0; i < class->size; i++) {
            if (class->members[i].sysex_addr_base > sysex_addr)
                return i - 1;
        }
	return (i - 1);
}

const char *libgieditor_get_desc(uint32_t sysex_addr) {
	midi_address *m_address = libgieditor_match_midi_address(sysex_addr);
	int i = match_class_member(sysex_addr, 0);
	if (i < 0) return NULL;
	return (m_address->class->members[i].name);
}

char **libgieditor_get_parents(uint32_t sysex_addr, int *num_parents) {
	midi_address *m_address = libgieditor_match_midi_address(sysex_addr);
	int depth, i, j;
	MidiClass *cur_class = m_address->class->parents[0];
	char **parents_array;
	
	*num_parents = 0;
	cur_class = m_address->class->parents[0];
	while(cur_class) {
	    cur_class = m_address->class->parents[*num_parents];
	    *num_parents = *num_parents + 1;
	}
	*num_parents = *num_parents - 1;

	if (*num_parents <= 0) return NULL;

	parents_array = allocate(char *, *num_parents);

	for (depth = *num_parents, j = 0; depth > 0; depth--, j++) {
	    i = match_class_member(sysex_addr, depth);
	    asprintf(&parents_array[j], "%s",
		    m_address->class->parents[depth - 1]->members[i].name);
	}

	return parents_array;
}

uint32_t libgieditor_get_sysex_size(uint32_t sysex_addr) {
	midi_address *m_address = libgieditor_match_midi_address(sysex_addr);
	if (!m_address) return -1;
	return m_address->sysex_size;
}

uint32_t libgieditor_get_sysex_value(uint8_t *data, uint32_t size) {
	int i, j;
	uint32_t retval = 0;
	j = size - 1;
	for (i = 0; i < size; i++, j--) {
		retval |= data[j] << (i*4);
	}
	return retval;
}

/* This function will block, returns the number of data bytes collected */
int libgieditor_listen_sysex_event(uint8_t *command_id, 
		uint32_t *address, uint8_t **data) {
	return sysex_listen_event(command_id, address, data);
}
