/* SysEx message handler
 *
 * Juno Gi Editor
 *
 * Copyright (C) 2012 Kim Taylor <kmtaylor@gmx.com>
 * Thanks to Claudio Matsuoka
 *
 *	This program is free software; you can redistribute it and/or modify it
 *	under the terms of the GNU General Public License as published by the
 *	Free Software Foundation version 2 of the License.
 * 
 *	This program is distributed in the hope that it will be useful, but
 *	WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *	General Public License for more details.
 * 
 *	You should have received a copy of the GNU General Public License along
 *	with this program; if not, write to the Free Software Foundation, Inc.,
 *	675 Mass Ave, Cambridge, MA 02139, USA.
 *
 * $Id: sysex.c,v 1.4 2012/06/08 09:58:20 kmtaylor Exp $
 */

#include <stdio.h>
#include <stdint.h>
#include <string.h>

#include <jack/jack.h>
#include <jack/midiport.h>

#include "libgieditor.h"

#define allocate(t, num) __common_allocate(sizeof(t) * num, "libgieditor")

#define MIDI_CMD_COMMON_SYSEX	    0xf0
#define MIDI_CMD_COMMON_SYSEX_END   0xf7
#define MIDI_ROLAND_ID		    0x41
#define MAX_SYSEX_SIZE		    512
#define SYSEX_DATA_OFFSET	    11
#define SYSEX_COMMAND_OFFSET	    6
#define SYSEX_ADDRESS_OFFSET	    7
#define SYSEX_NOT_DATA_BYTES	    13

typedef struct s_sysex_list *Sysex_list;
struct s_sysex_list {
	uint8_t		data[MAX_SYSEX_SIZE];
	int		size;
	Sysex_list	next;
};

/* Only access with midi_lock */
static Sysex_list sysex_list;

static jack_client_t *jack_client;
static jack_port_t *midi_in_port;
static jack_port_t *midi_out_port;

static pthread_mutex_t midi_lock = PTHREAD_MUTEX_INITIALIZER;
static pthread_cond_t data_ready = PTHREAD_COND_INITIALIZER;

static void add_sysex_event(uint8_t *data, int size) {
	Sysex_list cur_sysex;
	if (!sysex_list) {
	    sysex_list = allocate(struct s_sysex_list, 1);
	    cur_sysex = sysex_list;
	} else {
	    cur_sysex = sysex_list;
	    while (cur_sysex->next) cur_sysex = cur_sysex->next;
	    cur_sysex->next = allocate(struct s_sysex_list, 1);
	    cur_sysex = cur_sysex->next;
	}
	cur_sysex->next = NULL;
	cur_sysex->size = size;
	memcpy(cur_sysex->data, data, size);
}

static int jack_callback(jack_nframes_t nframes, void *arg) {
	jack_midi_event_t jack_midi_event;
	jack_nframes_t event_index = 0;

	pthread_mutex_lock(&midi_lock);

	void *midi_buf = jack_port_get_buffer(midi_in_port, nframes);
	
	while (jack_midi_event_get(&jack_midi_event, midi_buf, 
					event_index++) == 0) {
	    if (( jack_midi_event.buffer[0] == MIDI_CMD_COMMON_SYSEX ) &&
			( jack_midi_event.buffer[jack_midi_event.size - 1] == 
			  MIDI_CMD_COMMON_SYSEX_END )) {
		add_sysex_event(jack_midi_event.buffer,
				    jack_midi_event.size);
		pthread_cond_signal(&data_ready);
	    }
			
	}

	pthread_mutex_unlock(&midi_lock);

	return 0;
}

int sysex_init(const char *client_name) {
	jack_status_t jack_status;
	jack_client = 
		jack_client_open(client_name, JackNoStartServer, &jack_status);
	if (!jack_client) return -1;

	midi_in_port = jack_port_register(jack_client, "midi_in",
			JACK_DEFAULT_MIDI_TYPE,
			JackPortIsInput | JackPortIsTerminal, 0);
	
	midi_out_port = jack_port_register(jack_client, "midi_out",
			JACK_DEFAULT_MIDI_TYPE,
			JackPortIsOutput | JackPortIsTerminal, 0);
	
	if (!(midi_in_port || midi_out_port)) return -1;
	
	jack_set_process_callback(jack_client, jack_callback, 0);

	if (jack_activate(jack_client)) return -1;

	return 0;
}

int sysex_close(void) {
	return jack_client_close(jack_client);
}

int sysex_listen_event(uint8_t *command_id, 
		                uint32_t *address, uint8_t **data) {
	Sysex_list cur_sysex;
	int data_bytes = 0;

	pthread_mutex_lock(&midi_lock);

	while (sysex_list == NULL) {
	    pthread_cond_wait(&data_ready, &midi_lock);
	}

	cur_sysex = sysex_list;
	sysex_list = sysex_list->next;

	*command_id = cur_sysex->data[SYSEX_COMMAND_OFFSET];
	*address =  cur_sysex->data[SYSEX_ADDRESS_OFFSET]	<< 24 |
		    cur_sysex->data[SYSEX_ADDRESS_OFFSET+1]	<< 16 |
		    cur_sysex->data[SYSEX_ADDRESS_OFFSET+2]	<< 8 |
		    cur_sysex->data[SYSEX_ADDRESS_OFFSET+3];
	data_bytes = cur_sysex->size - SYSEX_NOT_DATA_BYTES;

	if (data_bytes > 0) {
	    *data = allocate(uint8_t, data_bytes);
	    memcpy(*data, cur_sysex->data + SYSEX_DATA_OFFSET, data_bytes);
	}
	free(cur_sysex);

        pthread_mutex_unlock(&midi_lock);

	return data_bytes;
}

int midi_sysex_recv(int buf_size, uint8_t *buf) {
	return 0;
}

static int checksum(int len, uint8_t *data)
{
	int i, sum;

	for (sum = i = 0; i < len; i++)
		sum = (sum + *data++) & 0x7f;

	return 0x80 - sum;
}

int sysex_send(int model, int dev_id, uint32_t addr, int len, uint8_t *data)
{
	static uint8_t buf[550];
	int sum, start;
//	struct fsex_id *id;
	int i, j;

	printf("dev_id: 0x%02x, addr: %08x, len: %d", dev_id, addr, len);
	if (len > 500) return -1;

	i = 0;
	buf[i++] = MIDI_CMD_COMMON_SYSEX;
	buf[i++] = MIDI_ROLAND_ID;	/* Roland ID */
	buf[i++] = dev_id;	/* Device ID */

//	for (j = 0; j < id->id_size; j++)
//		buf[i++] = id->id[j];

	buf[i++] = 0x12;	/* Command ID (DT1) */

	start = i;

	buf[i++] = (addr & 0xff000000) >> 24;
	buf[i++] = (addr & 0x00ff0000) >> 16;
	buf[i++] = (addr & 0x0000ff00) >> 8;
	buf[i++] = (addr & 0x000000ff);
	
	memcpy(buf + i, data, len);
	i += len;

	sum = checksum(len + 4, buf + start);

	buf[i++] = sum;
	buf[i++] = MIDI_CMD_COMMON_SYSEX_END;

//	midi_sysex_send(i, buf);
	return 0;
}

int recv_sysex(int model, int dev_id, uint32_t addr, int len, int dsize, uint8_t *data)
{
	static uint8_t buf[550];
	int sum, start, rlen;
//	struct fsex_id *id;
	int i, j;

	printf("dev_id: 0x%02x, addr: 0x%08x, len: %d", dev_id, addr, len);
	if (len > 500) return -1;

	i = 0;
	buf[i++] = MIDI_CMD_COMMON_SYSEX;
	buf[i++] = MIDI_ROLAND_ID;	/* Roland ID */
	buf[i++] = dev_id;	/* Device ID */

//	for (j = 0; j < id->id_size; j++)
//		buf[i++] = id->id[j];

	buf[i++] = 0x11;	/* Command ID (RQ1) */

	start = i;

	buf[i++] = (addr & 0xff000000) >> 24;
	buf[i++] = (addr & 0x00ff0000) >> 16;
	buf[i++] = (addr & 0x0000ff00) >> 8;
	buf[i++] = (addr & 0x000000ff);
	
	buf[i++] = (len & 0xff000000) >> 24;
	buf[i++] = (len & 0x00ff0000) >> 16;
	buf[i++] = (len & 0x0000ff00) >> 8;
	buf[i++] = (len & 0x000000ff);
	
	sum = checksum(i - start, buf + start);

	buf[i++] = sum;
	buf[i++] = MIDI_CMD_COMMON_SYSEX_END;

//	midi_sysex_send(i, buf);
	rlen = midi_sysex_recv(dsize, data);

	if (rlen < 2)
		return -1;

//	sum = checksum(rlen - 5 - id->id_size, data + 4 + id->id_size);

	printf("sum = %02x\n", sum);
	printf("data = %02x %02x\n", data[rlen - 2], data[rlen - 1]);

	if (sum != 0x80)
		return -1;

	return 0;
}
